#include "stdio.h"

#include "Client.h"
#include "Server.h"
#include "Board.h"
#include "Robot.h"
#include "Utils.h"
#include "Package.h"
#include "PackageList.h"

Client::Client(Server* server) : 
	m_board(0), m_robot(0), m_server(server), m_curPkgList(0)
{
}

Client::~Client()
{
	if (m_curPkgList)	
		delete m_curPkgList;

	if (m_board)
		delete m_board;

	if (m_robot)
		delete m_robot;
}

// --------------------------------------------------------
int Client::Run()
// --------------------------------------------------------
{
	int result = 0;

	if (Init())
	{
		//m_board->SaveMap("saved.map");

		// *************** Primary Turn Loop ***************
		bool ok = true;
		while (ok)
		{
			//m_board->PrintMap(m_robot->GetY()-1, m_robot->GetX()-1);

			ok = DoMove();
			if (ok)
			{
				ok = HandleReply();
			}
			//m_robot->Print();

			//getchar();
		}
		// *************************************************
		result = 1;
	}

	return result;
}

// --------------------------------------------------------
bool Client::DoMove()
// --------------------------------------------------------
{
	bool result = false;

	if (m_robot->PlanMove(m_board, &m_robotList, 0))
	{
		char move[MAX_LINE];
		if (m_robot->GetRequestLine(move))
		{
			result = m_server->SendCmd(move);
		}
	}

	return result;
}

// --------------------------------------------------------
bool Client::HandleReply()
// --------------------------------------------------------
{
	bool result = false;

	char reply[MAX_LINE];
	if (m_server->ReadLine(reply))
	{
		if (ParseReply(reply))
		{
			PackageList* pkgList = 0;
			if (ReadPackages(pkgList))
			{
				if (pkgList)
				{
					m_robot->OnFoundPackages(pkgList);
					delete pkgList;
				}
			}

			result = true;
		}
	}

	return result;
}

// --------------------------------------------------------
bool Client::ParseReply(char* line)
// --------------------------------------------------------
{
	char* p = line;
	Robot* robot = 0;
	int ok = true;
	bool nothing = false;
	bool foundPlayer = false;

	while ( (*p != 0) && (ok))
	{
		if (*p == '#')
		{
			// Found start of robot information
			int idRobot;
			
			++p;	// advance past '#'

			if (GetInt(p, &idRobot))
			{
				if (robot && nothing)
				{
					robot->OnNothing();
				}

				robot = m_robotList.FindRobot(idRobot);
				if (robot == 0)
				{
					printf("Created New Robot %d\n", idRobot);

					robot = m_robotList.CreateRobot(idRobot);
				}
				else
				{
					if (robot == m_robot)
					{
						// Make sure we find our player robot
						// in the reply, otherwise we died. Server
						// should also close connection
						foundPlayer = true;

						robot->m_callbacks = 0;
					}				
				}
				nothing = true;
			}
		}
		else
		{
			char action[MAX_LINE];
			int val;
			if (GetWord(p, action))
			{
				switch (action[0])
				{
					case 'N':
					case 'S':
					case 'E':
					case 'W':
						if (robot)
						{
							robot->OnMove(m_board, action[0]);
							nothing = false;
						}
						break;

					case 'P':
						if (GetInt(p, &val))
						{
							if (robot)
							{
								robot->OnPick(val);
								nothing = false;
							}
						}
						break;
					case 'D':
						if (GetInt(p, &val))
						{
							if (robot)
							{
								robot->OnDrop(val);
								nothing = false;
							}
						}
						break;
					case 'X':
						if (GetInt(p, &val))
						{
							if (robot)
							{
								robot->SetX(val);
								nothing = false;
							}
						}
						break;
					case 'Y':
						if (GetInt(p, &val))
						{
							if (robot)
							{
								robot->SetY(val);
								nothing = false;
							}
						}
						break;
					default:
						printf("UNEXPECTED action: %c\n", *p);
						ok = false;
						break;
				}
			}
		}
	}

	if (robot && nothing)
	{
		robot->OnNothing();
	}

	return foundPlayer;
}


// --------------------------------------------------------
bool Client::Init()
// --------------------------------------------------------
{
	bool result = false;

	if (m_server->Connect())
	{
		if (m_server->SendCmd("Player\n"))
		{
			if (ReadBoard())
			{
				if (ReadRobot())
				{
					if (ReadState())
					{
						if (ReadPackages(m_curPkgList))
						{
							result = true;	
						}
					}
				}
			}
		}
	}
	
	return result;
}

// --------------------------------------------------------
bool Client::ReadBoard()
// --------------------------------------------------------
{
	bool result = false;

	char line[MAX_LINE];
	if (m_server->ReadLine(line))
	{
		char* ptr = line;
		int width = 0;
		int height = 0;

		if (GetInt(ptr, &width))
		{
			if (GetInt(ptr, &height))
			{
				m_board = new Board(width, height);
				if (m_board)
				{
					// -------------- Read map data ----------------
					bool error = false;
					//for (int row = height-1; row >= 0; row--)
					for (int row = 0; row < height; row++)
					{
						if (!m_server->ReadLine(line))
						{
							error = true;
							break;
						}

						m_board->CopyRow(row, line);
					}

					result = !error;
				}
			}
		}

	}

	return result;
}

// --------------------------------------------------------
bool Client::ReadRobot()
// --------------------------------------------------------
{
	char line[MAX_LINE];
	if (m_server->ReadLine(line))
	{
		char* ptr = line;
		int id = 0;
		int maxCarryWeight = 0;
		int initialMoney = 0;

		if (GetInt(ptr, &id))
		{
			if (GetInt(ptr, &maxCarryWeight))
			{
				if (GetInt(ptr, &initialMoney))
				{
					m_robot = m_robotList.CreateRobot(id);
					if (m_robot)
					{
						m_robot->SetMaxCarryWeight(maxCarryWeight);
						m_robot->SetMoneyLeft(initialMoney);
					}
				}
			}
		}
	}
	
	return (m_robot != 0) ? true : false;
}

// --------------------------------------------------------
bool Client::ReadState()
// --------------------------------------------------------
{
	char line[MAX_LINE];
	bool result = m_server->ReadLine(line);
	if (result)
	{
		ParseReply(line);
	}

	return result;
}

// --------------------------------------------------------
bool Client::ReadPackages(PackageList*& pkgList)
// --------------------------------------------------------
{
	bool result = true;

	char line[MAX_LINE];
	if (m_server->ReadLine(line))
	{
		if (*line != 0)
		{
			char* ptr = line;
			bool ok = true;

			while (ok)
			{
				int id;
				int x;
				int y;
				int weight;

				if ((ok = GetInt(ptr, &id)))
				{
					if ((ok = GetInt(ptr, &x)))
					{
						if ((ok = GetInt(ptr, &y)))
						{
							if ((ok = GetInt(ptr, &weight)))
							{
								Package* pkg = new Package(id, x, y, weight);
								if (pkg)
								{
									if (pkgList != 0)
									{
										delete pkgList;
									}

									pkgList = new PackageList();
									if (pkgList)
									{
										pkgList->AddPackage(pkg);
									}	
									else 
										delete pkg;
								}
							}
						}
					}
				}
			}
			// else no packages
		}
	}

	return result;
}



